#include <cmath>
#include <iostream>
#include <stdexcept>

#include <controller/Controller.h>
#include <emulator/TrafficSign.hpp>

Controller::Controller(const State &state)
    : max_speed{state.getCar().getMaxSpeed(Gear::Forward)},
      min_speed{state.getCar().getMinSpeed()},
      starting_location{state.getCar().getLocation()},
      destination{state.getCar().getDestination().center},
      signs{state.getRoad().getSigns()},
      gas_pedal_force{Controller::default_gas_pedal_force},
      break_pedal_force{Controller::default_break_pedal_force} {};

LengthUnit Controller::get_deceleration_rate() const {
  return deceleration_rate.has_value() ? deceleration_rate.value()
                                       : Controller::default_deceleration_rate;
}

double Controller::compute_breaking_time(LengthUnit speed) const {
  LengthUnit dec_rate = this->break_pedal_force * this->get_deceleration_rate();
  if (dec_rate == 0) {
    throw std::runtime_error("zero deceleration rate");
  }
  double breaking_time = speed / dec_rate;
  return breaking_time;
}

LengthUnit Controller::compute_breaking_distance(LengthUnit speed) const {
  LengthUnit dec_rate = this->break_pedal_force * this->get_deceleration_rate();
  double breaking_time = this->compute_breaking_time(speed);
  LengthUnit breaking_dist =
      breaking_time * (speed - dec_rate * (breaking_time - 1) / 2);
  return breaking_dist;
}

void Controller::update_deceleration_rate(LengthUnit speed) {
  if (this->deceleration_rate.has_value() ||
      !this->previous_state.has_value()) {
    return;
  }
  LengthUnit prev_speed = this->previous_state->speed;
  if ((prev_speed <= 0) || (speed <= 0) || (prev_speed <= speed)) {
    return;
  }
  this->deceleration_rate.emplace((prev_speed - speed) /
                                  Controller::break_pedal_force);
}

std::optional<TrafficSign> Controller::get_last_sign(LengthUnit pos) const {
  auto max_sign_itr = this->signs.end();
  LengthUnit max_pos = std::numeric_limits<double>::min();

  for (auto sign_itr = this->signs.begin(); sign_itr != this->signs.end();
       ++sign_itr) {
    LengthUnit sign_pos = sign_itr->getLocation();
    if ((sign_pos < pos) && (sign_pos > max_pos)) {
      max_sign_itr = sign_itr;
      max_pos = sign_pos;
    }
  }

  if (max_sign_itr == this->signs.end()) {
    return std::nullopt;
  }
  return *max_sign_itr;
}

void Controller::modify_state(State &state) {
  Car &car = state.getCar();
  Location loc = car.getLocation();

  LengthUnit speed = car.getCurrentSpeed();
  this->update_deceleration_rate(speed);

  auto sign = this->get_last_sign(loc.x);
  if (sign.has_value()) {
    switch (sign->getType()) {
    case TrafficSign::Type::MinimumSpeedLimit:
      this->gas_pedal_force =
          std::max(Controller::default_gas_pedal_force, sign->getValue());
      break;
    case TrafficSign::Type::MaximumSpeedLimit:
      this->gas_pedal_force =
          std::min(Controller::default_gas_pedal_force, sign->getValue());
    }
  } else {
    this->gas_pedal_force = default_gas_pedal_force;
  }

  LengthUnit dist =
      std::hypot(this->destination.x - loc.x, this->destination.y - loc.y);
  LengthUnit breaking_dist =
      this->compute_breaking_distance(car.getCurrentSpeed());

  bool do_final_stop = (dist > breaking_dist + speed);

  if (do_final_stop) {
    car.setGear(Gear::Forward);
    car.pressBrakePedal(0.0);
    car.pressGasPedal(this->gas_pedal_force);
  } else {
    car.setGear(Gear::Neutral);
    car.pressGasPedal(0.0);
    car.pressBrakePedal(this->break_pedal_force);
  }

  this->previous_state.emplace(StateView{speed});
}