#include <controller/Controller.h>
#include <controller/Formatter.h>
#include <emulator/Emulator.hpp>

#include <iostream>
#include <optional>

class Application {
public:
  explicit Application(Mode mode) : mode{mode} {}

  void finish_initialization(const State &state) {
    if (this->lane_views.empty()) {
      for (const auto &lane : state.getRoad().getLanes()) {
        const auto loc = lane.getLocation();
        const auto dir = lane.getDirection();

        this->lane_views.push_back({
            static_cast<int>(std::round(loc.x)),
            static_cast<int>(std::round(loc.x + dir.dx)),
            static_cast<int>(std::round(loc.y)),
            static_cast<int>(std::round(loc.y + dir.dy)),
        });
      }
    }

    if (this->signs.empty()) {
      for (const auto &sign : state.getRoad().getSigns()) {
        this->signs.emplace_back(sign.getLocation());
      }
    }

    if (!this->controller.has_value()) {
      this->controller.emplace(Controller(state));
    }
  }

  bool run(State &state) {
    if (mode == Mode::HandsOn) {
      return false;
    }
    this->finish_initialization(state);
    this->controller->modify_state(state);
    this->draw(state);
    return !state.isDestination() && !state.isCollision();
  }

  void draw(State &state) const {
    const auto &car = state.getCar();
    const auto location = car.getLocation();
    Formatter formatter(location);

    formatter.format_lanes(this->lane_views);

    std::vector<Location> other_cars_locations;
    for (const auto &vehicle : state.getVisibleVehicles()) {
      other_cars_locations.emplace_back(vehicle.getLocation());
    }
    formatter.format_other_cars(other_cars_locations);
    formatter.format_my_params(car.getLocation(), car.getCurrentSpeed());
    formatter.format_signs(this->signs);
    formatter.format_destination(car.getDestination().center);
    formatter.format_my_car();

    formatter.print(true);
  }

private:
  const Mode mode;

private:
  std::vector<LaneView> lane_views;
  std::vector<LengthUnit> signs;

  std::optional<Controller> controller;
};

int main(int argc, char **argv) {
  try {
    Emulator emulator(argc, argv);
    Application app(emulator.getMode());
    emulator.run([&app](State &state) { return app.run(state); });
    return 0;
  } catch (const std::runtime_error &error) {
    std::cerr << error.what() << std::endl;
    return 1;
  }
}
